/*
 * MakeUNIF v1.0, by Quietust
 *
 * Makefile structure
 *	MAKEUNIF		- File identification tag - MUST be the first line in the file
 *	filename.unif		- Name of the .UNIF file to generate
 *	MAPR NES-NROM-128	- Specify the board name to use
 *	NAME Game title		- Specify a game title to store
 *	TVCI n			- Specify 0 for NTSC, 1 for PAL, 2 for both
 *	MIRR n			- Specify mirroring (0=Horiz, 1=Vert, 2=OneScreen-L, 3=OneScreen-H, 4=FourScreen, 5=Dynamic)
 *	BATR n			- Specify the presence of a battery (value is ignored)
 *	READ readme.txt		- Embed a segment of text within the UNIF file
 *	PRG0 filename.obj	- Specify a PRG ROM segment (can specify additional segments using PRG1 thru PRGF)
 *	CHR0 filename.chr	- Specify a CHR ROM segment (optional; can specify additional segments using CHR1 thru CHRF)
 *
 * The following UNIF blocks are unsupported:
 *	DINF - Specifies information on when a game was dumped, very poorly structured block
 *	CTRL - Deemed unnecessary - games can use whatever controllers they want
 *	PCKn/CCKn - Can be added by other tools
 *	VROR - Deemed unnecessary - if you want to use RAM instead of ROM, invent a new board name
 */

#include <stdio.h>

#define	MAXLEN	256

char path[MAXLEN];
void	SetPath (const char *filename)
{
	int i;
	strcpy(path,filename);
	for (i = strlen(path) - 1; i >= 0; i--)
		if (path[i] == '\\')
			break;
	path[i+1] = 0;
}

char fname[MAXLEN];
char *	GetPath (const char *filename)
{
	strcpy(fname,path);
	strcat(fname,filename);
	return fname;
}

int	SplitLine (const char *line, char *header, char *data)
{
	int i;
	int j = 0;
	int hoff = 0, doff = 0;
	for (i = 0; i < strlen(line); i++)
	{
		if (j == 0)
		{
			if (line[i] == ' ')
				j = 1;
			else	header[hoff++] = line[i];
		}
		else	data[doff++] = line[i];
	}
	header[hoff] = 0;
	data[doff] = 0;
	if (header[hoff-1] == '\n')
		return 0;
	if (data[doff-1] == '\n')
	{
		data[doff-1] = 0;
		return 1;
	}
	else	return 0;
}

int main (int argc, char **argv)
{
	FILE *makeinfo;
	FILE *output;
	char data[MAXLEN];
	int i;
	if (argc < 2)
	{
		fprintf(stderr,"Syntax: %s <makefile>\n",argv[0]);
		return 1;
	}
	makeinfo = fopen(argv[1],"rt");
	if (makeinfo == NULL)
	{
		fprintf(stderr,"Unable to open makefile!\n");
		return 1;
	}
	SetPath(argv[1]);
	fgets(data,MAXLEN,makeinfo);
	if (strcmp(data,"MAKEUNIF\n"))
	{
		fclose(makeinfo);
		fprintf(stderr,"Specified file is not a valid makefile!\n");
		return 1;
	}
	fgets(data,MAXLEN,makeinfo);
	if (data[strlen(data)-1] != '\n')
	{
		fclose(makeinfo);
		fprintf(stderr,"Filename too long! Increase MAXLEN and recompile.\n");
		return 1;
	}
	else	data[strlen(data)-1] = 0;
	output = fopen(GetPath(data),"wb");
	if (output == NULL)
	{
		fclose(makeinfo);
		fprintf(stderr,"Unable to open file %s for output!\n",data);
		return 1;
	}
	printf("Opened file %s for output.\n",data);
	{	// write UNIF header
		unsigned long ver = 7;
		fwrite("UNIF",4,1,output);
		fwrite(&ver,4,1,output);
		ver = 0;
		fwrite(&ver,4,1,output);
		fwrite(&ver,4,1,output);
		fwrite(&ver,4,1,output);
		fwrite(&ver,4,1,output);
		fwrite(&ver,4,1,output);
		fwrite(&ver,4,1,output);
		printf("Writing header...\n");
	}

	while (fgets(data,MAXLEN,makeinfo))
	{
		char blk[MAXLEN];
		char args[MAXLEN];
		unsigned long blklen;
		unsigned char tbyte;
		if (feof(makeinfo))
			break;
		if (!SplitLine(data,blk,args))
		{
			fclose(makeinfo);
			fclose(output);
			fprintf(stderr,"Error decoding line: %s\n",data);
			return 1;
		}
		if ((!memcmp(blk,"MAPR",4)) || (!memcmp(blk,"NAME",4)))
		{
			blklen = strlen(args)+1;
			printf("Storing %s as '%s'\n",blk,args);
			fwrite(blk,4,1,output);
			fwrite(&blklen,4,1,output);
			fwrite(args,blklen,1,output);
		}
		else if ((!memcmp(blk,"TVCI",4)) || (!memcmp(blk,"MIRR",4)) || (!memcmp(blk,"BATR",4)))
		{
			sscanf(args,"%d",&tbyte);
			blklen = 1;
			printf("Storing %s as %i\n",blk,tbyte);
			fwrite(blk,4,1,output);
			fwrite(&blklen,4,1,output);
			fwrite(&tbyte,1,1,output);
		}
		else if ((!memcmp(blk,"READ",4)) || (!memcmp(blk,"PRG",3)) || (!memcmp(blk,"CHR",3)))
		{
			FILE *rom = fopen(GetPath(args),"rb");
			if (rom == NULL)
			{
				fprintf(stderr,"Error writing block %s - unable to open file '%s'!\n",blk,args);
				continue;
			}
			fseek(rom,0,SEEK_END);
			blklen = ftell(rom);
			fseek(rom,0,SEEK_SET);
			if (!memcmp(blk,"READ",4))
				blklen++;
			printf("Storing %s as from file %s (%i bytes)\n",blk,args,blklen);
			fwrite(blk,4,1,output);
			fwrite(&blklen,4,1,output);
			if (!memcmp(blk,"READ",4))
				blklen--;
			for (i = 0; i < blklen; i++)
			{
				fread(&tbyte,1,1,rom);
				fwrite(&tbyte,1,1,output);
			}
			if (!memcmp(blk,"READ",4))
			{
				tbyte = 0;
				fwrite(&tbyte,1,1,output);
			}
			fclose(rom);
		}
		else	fprintf(stderr,"Unrecognized directive %s encountered, skipping...\n",blk);
	}
	fclose(makeinfo);
	fclose(output);
	printf("Done!\n");
	return 0;
}